"""
Test the Armasec convenience class.
"""

import asgi_lifespan
import fastapi
import httpx
import pendulum
import pytest
import starlette
from fastapi import HTTPException
from plummet import frozen_time

from armasec import Armasec, TokenSecurity


@pytest.fixture
async def app():
    """
    Provides an instance of a FastAPI app that is to be used only for testing purposes.
    """
    return fastapi.FastAPI()


@pytest.fixture
async def build_secure_endpoint(app):
    """
    Provides a method that dynamically builds a GET route on the app that requires a header with a
    valid token that includes the supplied scopes.
    """

    def _helper(path: str, security: TokenSecurity):
        """
        Adds a route onto the app at the provide path that is secured by the provided Armasec
        instance.
        """

        @app.get(path, dependencies=[fastapi.Depends(security)])
        async def _():
            return dict(good="to go")

    return _helper


@pytest.fixture
async def client(app):
    """
    Provides a FastAPI client against which httpx requests can be made. Includes a "/secure"
    endpoint that requires auth via the TokenSecurity injectable.
    """
    async with asgi_lifespan.LifespanManager(app):
        async with httpx.AsyncClient(transport=httpx.ASGITransport(app=app), base_url="http://armasec-test") as client:
            yield client


@frozen_time("2021-09-20 11:02:00")
async def test_lockdown__with_no_scopes(
    mock_openid_server,
    rs256_domain,
    build_rs256_token,
    build_secure_endpoint,
    app,
    client,
):
    """
    Test that lockdown works correctly when supplied no scopes.
    """

    armasec = Armasec(domain=rs256_domain, audience="https://this.api")

    exp = pendulum.parse("2021-09-21 11:02:00", tz="UTC")
    token = build_rs256_token(claim_overrides=dict(sub="me", exp=exp.timestamp()))
    build_secure_endpoint("/secured-no-scopes", armasec.lockdown())

    response = await client.get("/secured-no-scopes", headers={"Authorization": f"bearer {token}"})
    assert response.status_code == starlette.status.HTTP_200_OK

    response = await client.get("/secured-no-scopes")
    assert response.status_code == starlette.status.HTTP_401_UNAUTHORIZED


@frozen_time("2021-09-20 11:02:00")
async def test_lockdown__with_scopes(
    mock_openid_server,
    rs256_domain,
    build_rs256_token,
    build_secure_endpoint,
    app,
    client,
):
    """
    Test that lockdown works correctly when supplied with scopes.
    """

    armasec = Armasec(domain=rs256_domain, audience="https://this.api")
    build_secure_endpoint("/secured-with-scopes", armasec.lockdown("read:more"))

    good_token = build_rs256_token(
        claim_overrides=dict(
            sub="me",
            exp=pendulum.parse("2021-09-21 11:02:00", tz="UTC").timestamp(),
            permissions=["read:more"],
        ),
    )
    response = await client.get(
        "/secured-with-scopes",
        headers={"Authorization": f"bearer {good_token}"},
    )
    assert response.status_code == starlette.status.HTTP_200_OK

    bad_token = build_rs256_token(
        claim_overrides=dict(
            sub="me",
            exp=pendulum.parse("2021-09-21 11:02:00", tz="UTC").timestamp(),
        ),
    )
    response = await client.get(
        "/secured-with-scopes",
        headers={"Authorization": f"bearer {bad_token}"},
    )
    assert response.status_code == starlette.status.HTTP_403_FORBIDDEN


@frozen_time("2021-09-22 22:54:00")
async def test_lockdown__with_all_scopes(
    mock_openid_server,
    rs256_domain,
    build_rs256_token,
    build_secure_endpoint,
    app,
    client,
):
    """
    Test that lockdown works correctly requiring all scopes.
    """

    armasec = Armasec(domain=rs256_domain, audience="https://this.api")
    build_secure_endpoint("/secured-with-scopes", armasec.lockdown_all("read:one", "read:more"))

    good_token = build_rs256_token(
        claim_overrides=dict(
            sub="me",
            exp=pendulum.parse("2021-09-23 22:54:00", tz="UTC").timestamp(),
            permissions=["read:one", "read:more"],
        ),
    )
    response = await client.get(
        "/secured-with-scopes",
        headers={"Authorization": f"bearer {good_token}"},
    )
    assert response.status_code == starlette.status.HTTP_200_OK

    bad_token = build_rs256_token(
        claim_overrides=dict(
            sub="me",
            exp=pendulum.parse("2021-09-23 22:54:00", tz="UTC").timestamp(),
            permissions=["read:one"],
        ),
    )
    response = await client.get(
        "/secured-with-scopes",
        headers={"Authorization": f"bearer {bad_token}"},
    )
    assert response.status_code == starlette.status.HTTP_403_FORBIDDEN


@frozen_time("2021-09-22 22:54:00")
async def test_lockdown__with_some_scopes(
    mock_openid_server,
    rs256_domain,
    build_rs256_token,
    build_secure_endpoint,
    app,
    client,
):
    """
    Test that lockdown works correctly requiring some scopes.
    """

    armasec = Armasec(domain=rs256_domain, audience="https://this.api")
    build_secure_endpoint("/secured-with-scopes", armasec.lockdown_some("read:one", "read:more"))

    good_token = build_rs256_token(
        claim_overrides=dict(
            sub="me",
            exp=pendulum.parse("2021-09-23 22:54:00", tz="UTC").timestamp(),
            permissions=["read:one"],
        ),
    )
    response = await client.get(
        "/secured-with-scopes",
        headers={"Authorization": f"bearer {good_token}"},
    )
    assert response.status_code == starlette.status.HTTP_200_OK

    bad_token = build_rs256_token(
        claim_overrides=dict(
            sub="me",
            exp=pendulum.parse("2021-09-23 22:54:00", tz="UTC").timestamp(),
            permissions=[],
        ),
    )
    response = await client.get(
        "/secured-with-scopes",
        headers={"Authorization": f"bearer {bad_token}"},
    )
    assert response.status_code == starlette.status.HTTP_403_FORBIDDEN


@frozen_time("2021-09-20 11:02:00")
async def test_lockdown__with_two_domains__secondary_one_is_mocked(
    mock_openid_server,
    rs256_domain_config,
    rs256_secondary_domain_config,
    build_rs256_token,
    build_secure_endpoint,
    app,
    client,
):
    """
    Test that lockdown works correctly when supplied two domains to the Armasec class. The secondary
    input domain is the one to match the tokens and authenticate the incoming token.
    """

    armasec = Armasec(domain_configs=[rs256_secondary_domain_config, rs256_domain_config])

    exp = pendulum.parse("2021-09-21 11:02:00", tz="UTC")
    token = build_rs256_token(
        claim_overrides=dict(sub="me", exp=exp.timestamp(), permissions=["read:stuff"])
    )
    build_secure_endpoint("/secured-no-scopes", armasec.lockdown("read:stuff"))

    response = await client.get("/secured-no-scopes", headers={"Authorization": f"bearer {token}"})
    assert response.status_code == starlette.status.HTTP_200_OK

    response = await client.get("/secured-no-scopes")
    assert response.status_code == starlette.status.HTTP_401_UNAUTHORIZED


@frozen_time("2021-09-20 11:02:00")
async def test_lockdown__with_two_domains__first_one_is_mocked(
    mock_openid_server,
    rs256_domain_config,
    rs256_secondary_domain_config,
    build_rs256_token,
    build_secure_endpoint,
    app,
    client,
):
    """
    Test that lockdown works correctly when supplied two domains to the Armasec class. The first
    input domain is the one to match the tokens and authenticate the incoming token.
    """

    armasec = Armasec(domain_configs=[rs256_domain_config, rs256_secondary_domain_config])

    exp = pendulum.parse("2021-09-21 11:02:00", tz="UTC")
    token = build_rs256_token(
        claim_overrides=dict(sub="me", exp=exp.timestamp(), permissions=["read:stuff"])
    )
    build_secure_endpoint("/secured-no-scopes", armasec.lockdown("read:stuff"))

    response = await client.get("/secured-no-scopes", headers={"Authorization": f"bearer {token}"})
    assert response.status_code == starlette.status.HTTP_200_OK

    response = await client.get("/secured-no-scopes")
    assert response.status_code == starlette.status.HTTP_401_UNAUTHORIZED


@frozen_time("2021-09-20 11:02:00")
async def test_lockdown__with_two_domains__check_if_passing_domain_takes_precedence_over_the_domain_configs_list(  # noqa
    mock_openid_server,
    rs256_domain,
    rs256_secondary_domain_config,
    build_rs256_token,
    build_secure_endpoint,
    app,
    client,
):
    """
    Test if the Armasec class works properly when both domain_configs and domain are inputted. The
    expected behaviour is the domain itself to take precedence over the domain config list.
    """

    armasec = Armasec(
        domain_configs=[rs256_secondary_domain_config],
        domain=rs256_domain,
        audience="https://this.api",
    )

    exp = pendulum.parse("2021-09-21 11:02:00", tz="UTC")
    token = build_rs256_token(
        claim_overrides=dict(sub="me", exp=exp.timestamp(), permissions=["read:stuff"])
    )
    build_secure_endpoint("/secured-no-scopes", armasec.lockdown("read:stuff"))

    response = await client.get("/secured-no-scopes", headers={"Authorization": f"bearer {token}"})
    assert response.status_code == starlette.status.HTTP_200_OK

    response = await client.get("/secured-no-scopes")
    assert response.status_code == starlette.status.HTTP_401_UNAUTHORIZED


@frozen_time("2021-09-20 11:02:00")
async def test_lockdown__check_if_its_possible_to_pass_only_one_domain_config_as_a_list(
    mock_openid_server,
    rs256_domain_config,
    build_rs256_token,
    build_secure_endpoint,
    app,
    client,
):
    """
    Test if the Armasec class works properly only when the domain config list is provided.
    """

    armasec = Armasec(domain_configs=[rs256_domain_config])

    exp = pendulum.parse("2021-09-21 11:02:00", tz="UTC")
    token = build_rs256_token(
        claim_overrides=dict(sub="me", exp=exp.timestamp(), permissions=["read:stuff"])
    )
    build_secure_endpoint("/secured-no-scopes", armasec.lockdown("read:stuff"))

    response = await client.get("/secured-no-scopes", headers={"Authorization": f"bearer {token}"})
    assert response.status_code == starlette.status.HTTP_200_OK

    response = await client.get("/secured-no-scopes")
    assert response.status_code == starlette.status.HTTP_401_UNAUTHORIZED


@frozen_time("2021-09-20 11:02:00")
def test_armasec__no_domain_was_inputted():
    """
    Test if error is raised when neither domain nor domain_configs are provided to Armasec's class.
    """

    with pytest.raises(HTTPException) as err:
        Armasec()

    assert err.value.status_code == starlette.status.HTTP_422_UNPROCESSABLE_ENTITY
    assert err.value.detail == "No domain was input."


@frozen_time("2021-09-20 11:02:00")
@pytest.mark.parametrize(
    "key_value_pairs_to_match",
    [
        {"custom-key": "custom-value"},
        {"custom-key-1": "custom-value-1", "custom-key-2": "custom-value-2"},
        {"custom-key": ["custom-value"]},
        {"custom-key": ["custom-value-1", "custom-value-2"]},
        {"custom-key": True},
        {"custom-key": 1},
        {"custom-key": 1.0},
        {"custom-key": {"another-custom-key": "custom-value"}},
    ],
)
async def test_lockdown__test_match_keys__check_if_can_authorize(
    key_value_pairs_to_match,
    mock_openid_server,
    rs256_domain_config,
    build_rs256_token,
    build_secure_endpoint,
    app,
    client,
):
    """
    Test that lockdown works correctly when supplied two domains to the Armasec class. The first
    input domain is the one to match the tokens and authenticate the incoming token.
    """
    rs256_domain_config.match_keys = key_value_pairs_to_match

    armasec = Armasec(domain_configs=[rs256_domain_config])

    exp = pendulum.parse("2021-09-21 11:02:00", tz="UTC")
    token = build_rs256_token(
        claim_overrides=dict(
            sub="me", exp=exp.timestamp(), permissions=["read:stuff"], **key_value_pairs_to_match
        )
    )
    build_secure_endpoint("/secured-no-scopes", armasec.lockdown("read:stuff"))

    response = await client.get("/secured-no-scopes", headers={"Authorization": f"bearer {token}"})
    assert response.status_code == starlette.status.HTTP_200_OK

    response = await client.get("/secured-no-scopes")
    assert response.status_code == starlette.status.HTTP_401_UNAUTHORIZED


@frozen_time("2021-09-20 11:02:00")
@pytest.mark.parametrize(
    "key_value_pairs_to_match",
    [
        {"custom-key": "custom-value"},
        {"custom-key-1": "custom-value-1", "custom-key-2": "custom-value-2"},
        {"custom-key": ["custom-value"]},
        {"custom-key": ["custom-value-1", "custom-value-2"]},
        {"custom-key": True},
        {"custom-key": 1},
        {"custom-key": 1.0},
        {"custom-key": {"another-custom-key": "custom-value"}},
    ],
)
async def test_lockdown__test_match_keys__check_if_cannot_authorize(
    key_value_pairs_to_match,
    mock_openid_server,
    rs256_domain_config,
    build_rs256_token,
    build_secure_endpoint,
    app,
    client,
):
    """
    Test that lockdown works correctly when supplied two domains to the Armasec class. The first
    input domain is the one to match the tokens and authenticate the incoming token.
    """
    rs256_domain_config.match_keys = key_value_pairs_to_match

    armasec = Armasec(domain_configs=[rs256_domain_config])

    exp = pendulum.parse("2021-09-21 11:02:00", tz="UTC")
    token = build_rs256_token(
        claim_overrides=dict(sub="me", exp=exp.timestamp(), permissions=["read:stuff"])
    )
    build_secure_endpoint("/secured-no-scopes", armasec.lockdown("read:stuff"))

    response = await client.get("/secured-no-scopes", headers={"Authorization": f"bearer {token}"})
    assert response.status_code == starlette.status.HTTP_403_FORBIDDEN



import pendulum
import pytest
from plummet import frozen_time

from armasec.exceptions import AuthenticationError
from armasec.schemas.jwks import JWKs
from armasec.token_decoder import TokenDecoder
from armasec.token_manager import TokenManager


@pytest.fixture
def manager(rs256_openid_config, rs256_jwk):
    """
    This fixture provides a TokenManager configured with armasec's pytest_extension jwks and
    a token decoder using the same.
    """
    jwks = JWKs(keys=[rs256_jwk])
    decoder = TokenDecoder(jwks)
    return TokenManager(rs256_openid_config, decoder)


@frozen_time("2021-08-12 16:38:00")
def test_unpack_token_from_header__success(manager):
    """
    This test verifies that the ``unpack_token_from_header()`` method can successfully unpack a jwt
    from a header mapping where it is embedded with a scheme marker (like 'bearer').
    """
    headers = {"Authorization": "bearer dummy-token"}
    assert manager.unpack_token_from_header(headers) == "dummy-token"


def test_unpack_token_from_header__fail_when_auth_header_not_found(manager):
    """
    This test ensures that ``unpack_token_from_header()`` raises an AuthenticationError when an auth
    header is not found in header dict.
    """
    with pytest.raises(AuthenticationError, match="Could not find auth header"):
        manager.unpack_token_from_header(dict())


def test_unpack_token_from_header__fail_on_no_scheme_prefix(manager):
    """
    This test ensures that ``unpack_token_from_header()`` raises an AuthenciationError when an auth
    header lacks the correct scheme prefix before the jwt.
    """
    with pytest.raises(AuthenticationError, match="Could not extract scheme"):
        manager.unpack_token_from_header(dict(Authorization="xxxxxxxxxxxx"))


def test_unpack_token_from_header__fail_when_no_token(manager):
    """
    This test ensures that ``unpack_token_from_header()`` raises an AuthenciationError when an auth
    header lacks the embedded jwt.
    """
    with pytest.raises(AuthenticationError, match="Could not extract scheme"):
        manager.unpack_token_from_header(dict(Authorization="bearer "))


def test_unpack_token_from_header__fail_for_invalid_scheme(manager):
    """
    This  test ensures that ``unpack_token_from_header()`` raises an AutenticationError when an auth
    header lacks a maching scheme.
    """
    with pytest.raises(AuthenticationError, match="Invalid auth scheme"):
        manager.unpack_token_from_header(dict(Authorization="carrier xxxxxxxxxxxx"))


@frozen_time("2021-09-16 20:56:00")
def test_extract_token_payload__success(manager, build_rs256_token):
    """
    This test verifies that the ``extract_token_payload()`` method successfully decodes a jwt and
    produces a TokenPayload instance from the contents.
    """
    exp = pendulum.parse("2021-09-17 20:56:00", tz="UTC")
    token = build_rs256_token(
        claim_overrides=dict(sub="me", permissions=["read:all"], exp=exp.int_timestamp),
    )
    token_payload = manager.extract_token_payload({"Authorization": f"bearer {token}"})
    assert token_payload.sub == "me"
    assert token_payload.permissions == ["read:all"]
    assert token_payload.expire == exp


@frozen_time("2021-09-16 20:56:00")
def test_extract_token_payload__with_audience_succeeds(
    build_rs256_token, rs256_openid_config, rs256_jwk
):
    """
    This test verifies that the ``extract_token_payload()`` method successfully decodes a jwt that
    includes an audience claim when the TokenManager is initialized with an audience.
    """
    jwks = JWKs(keys=[rs256_jwk])
    decoder = TokenDecoder(jwks)
    manager = TokenManager(rs256_openid_config, decoder, audience="some-audience")
    exp = pendulum.parse("2021-09-17 20:56:00", tz="UTC")
    token = build_rs256_token(
        claim_overrides=dict(
            sub="me",
            exp=exp.timestamp(),
            audience="some-audience",
        ),
    )
    token_payload = manager.extract_token_payload({"Authorization": f"bearer {token}"})
    assert token_payload.sub == "me"


@frozen_time("2021-09-16 20:56:00")
def test_extract_token_payload__fails_without_audience(
    build_rs256_token, rs256_openid_config, rs256_jwk
):
    """
    This test verifies that the ``extract_token_payload()`` method raises an exception when the
    jwt carries an audience claim and the TokenManager is not initialized with an audience.
    """
    jwks = JWKs(keys=[rs256_jwk])
    decoder = TokenDecoder(jwks)
    manager = TokenManager(rs256_openid_config, decoder)
    exp = pendulum.parse("2021-09-17 20:56:00", tz="UTC")
    token = build_rs256_token(
        claim_overrides=dict(
            sub="me",
            exp=exp.timestamp(),
            aud="some-audience",
        ),
    )
    with pytest.raises(AuthenticationError):
        manager.extract_token_payload({"Authorization": f"bearer {token}"})


@frozen_time("2021-09-16 20:56:00")
def test_extract_token_payload__fails_with_bad_audience(
    build_rs256_token, rs256_openid_config, rs256_jwk
):
    """
    This test verifies that the ``extract_token_payload()`` method raises an exception when the
    jwt's audience does not match the audience of the TokenManager.
    """
    jwks = JWKs(keys=[rs256_jwk])
    decoder = TokenDecoder(jwks)
    manager = TokenManager(rs256_openid_config, decoder, audience="other-audience")
    exp = pendulum.parse("2021-09-17 20:56:00", tz="UTC")
    token = build_rs256_token(
        claim_overrides=dict(
            sub="me",
            exp=exp.timestamp(),
            aud="some-audience",
        ),
    )
    with pytest.raises(AuthenticationError):
        manager.extract_token_payload({"Authorization": f"bearer {token}"})



import pytest

from armasec.schemas import DomainConfig


@pytest.fixture()
def rs256_secondary_domain_config():
    """
    Return the DomainConfig model for the default rs256 domain.
    """
    return DomainConfig(domain="secondary.armasec.dev", audience="https://this.api")



from armasec import Armasec
from fastapi import FastAPI, Depends


app = FastAPI()
armasec = Armasec(
    domain="localhost:8080/realms/master",
    audience="http://keycloak.local",
    use_https=False,
    payload_claim_mapping=dict(permissions="resource_access.armasec_tutorial.roles"),
    debug_logger=print,
    debug_exceptions=True,
)

@app.get("/stuff", dependencies=[Depends(armasec.lockdown("read:stuff"))])
async def check_access():
    return dict(message="Successfully authenticated!")



import os

from armasec import Armasec
from armasec.schemas import DomainConfig
from fastapi import FastAPI, Depends


app = FastAPI()
armasec = Armasec(
    domain_configs=[
        DomainConfig(
            domain=os.environ.get("ARMASEC_DOMAIN_1"),
            audience=os.environ.get("ARMASEC_AUDIENCE_1"),
            match_keys={"dummy-key": "this value must be in the token"},
        ),
        DomainConfig(
            domain=os.environ.get("ARMASEC_DOMAIN_2"),
            audience=os.environ.get("ARMASEC_AUDIENCE_2"),
        ),
    ],
)

@app.get("/stuff", dependencies=[Depends(armasec.lockdown("read:stuff"))])
async def check_access():
    return dict(message="Successfully authenticated!")



"""
Test the openid_config_loader module.
"""

from unittest import mock

import httpx
import pytest
import respx
import starlette

from armasec.exceptions import AuthenticationError
from armasec.openid_config_loader import OpenidConfigLoader
from armasec.schemas import JWKs


def test_config__is_lazy_loaded(rs256_domain, mock_openid_server):
    """
    Verify that the config property lazy loads from the openid-configuration route.
    """
    OpenidConfigLoader(rs256_domain)
    assert not mock_openid_server.openid_config_route.called


def test_build_openid_config_url__default_protocol():
    """
    Verify that the openid config url is built correctly using the default protocol.
    """
    assert (
        OpenidConfigLoader.build_openid_config_url("my.domain")
        == "https://my.domain/.well-known/openid-configuration"
    )


def test_build_openid_config_url__with_other_protocol():
    """
    Verify that the openid config url is built correctly using a non-default protocol.
    """
    assert (
        OpenidConfigLoader.build_openid_config_url("my.domain", use_https=False)
        == "http://my.domain/.well-known/openid-configuration"
    )


def test__load_openid_resource__success(mock_openid_server):
    """
    Verify that the helper method can fetch json data via a GET call to a url.
    """
    loader = OpenidConfigLoader("my.domain")
    with respx.mock:
        route = respx.get("https://my.domain/blah")
        route.return_value = httpx.Response(
            starlette.status.HTTP_200_OK,
            json=dict(foo="bar"),
        )
        assert loader._load_openid_resource("https://my.domain/blah") == dict(foo="bar")


def test__load_openid_resource__with_http(mock_openid_server):
    """
    Verify that the helper method can fetch json data via a GET call to a url.
    """
    loader = OpenidConfigLoader("my.domain", use_https=False)
    with respx.mock:
        route = respx.get("http://my.domain/blah")
        route.return_value = httpx.Response(
            starlette.status.HTTP_200_OK,
            json=dict(foo="bar"),
        )
        assert loader._load_openid_resource("http://my.domain/blah") == dict(foo="bar")


def test__load_openid_resource__fails_on_failed_request(mock_openid_server):
    """
    Verify that the helper method throws an exception if the GET request fails.
    """
    loader = OpenidConfigLoader("my.domain")
    with mock.patch("httpx.get", side_effect=Exception("BOOM!")):
        with pytest.raises(AuthenticationError, match="Call to url .* failed"):
            loader._load_openid_resource("https://my.domain/blah")


def test__load_openid_resource__fails_on_bad_status_code(mock_openid_server):
    """
    Verify that the helper method throws an exception if the response status code is not OK.
    """
    loader = OpenidConfigLoader("my.domain")
    with respx.mock:
        route = respx.get("https://my.domain/blah")
        route.return_value = httpx.Response(starlette.status.HTTP_400_BAD_REQUEST)
        with pytest.raises(AuthenticationError, match="Didn't get a success status code"):
            loader._load_openid_resource("https://my.domain/blah")


def test_config__success(mock_openid_server, rs256_openid_config, rs256_domain):
    """
    Verify that the config property successfully loads a config from the server. Also verify that
    it will re-use the lazy-loaded config on subsequent calls.
    """
    loader = OpenidConfigLoader(rs256_domain)
    config = loader.config
    assert config.issuer == rs256_openid_config.issuer
    assert config.jwks_uri == rs256_openid_config.jwks_uri
    assert mock_openid_server.openid_config_route.called

    loader.config
    assert mock_openid_server.openid_config_route.call_count == 1


def test_config__fail_on_invalid_config(mock_openid_server, rs256_openid_config, rs256_domain):
    """
    Verify that the config property throws an exception if the loaded config was invalid.
    """
    loader = OpenidConfigLoader(rs256_domain)
    mock_openid_server.openid_config_route.return_value = httpx.Response(
        starlette.status.HTTP_200_OK,
        json=dict(bad="data"),
    )
    with pytest.raises(AuthenticationError, match="openid config data was invalid"):
        loader.config


def test_jwks__success(mock_openid_server, rs256_jwk, rs256_domain):
    """
    Verify that the jwks property successfully loads jwks from the server. Also verify that
    it will re-use the lazy-loaded jwks on subsequent calls.
    """
    loader = OpenidConfigLoader(rs256_domain)
    jwks = loader.jwks
    assert jwks == JWKs(keys=[rs256_jwk])
    assert mock_openid_server.jwks_route.called

    loader.jwks
    assert mock_openid_server.jwks_route.call_count == 1


def test_jwks__fail_on_invalid_jwks(mock_openid_server, rs256_jwk, rs256_domain):
    """
    Verify that the config property throws an exception if the loaded config was invalid.
    """
    loader = OpenidConfigLoader(rs256_domain)
    mock_openid_server.jwks_route.return_value = httpx.Response(
        starlette.status.HTTP_200_OK,
        json=dict(bad="data"),
    )
    with pytest.raises(AuthenticationError, match="jwks data was invalid"):
        loader.jwks



import json
from typing import Dict, Optional

import httpx
import pydantic
import pytest

from armasec_cli.exceptions import Abort
from armasec_cli.client import _deserialize_request_model, make_request


DEFAULT_DOMAIN = "https://dummy-domain.com"


@pytest.fixture
def dummy_client():
    def _helper(
        base_url: str = DEFAULT_DOMAIN,
        headers: Optional[Dict] = None,
    ) -> httpx.Client:
        if headers is None:
            headers = dict()

        return httpx.Client(base_url=base_url, headers=headers)

    return _helper


class DummyResponseModel(pydantic.BaseModel):
    foo: int
    bar: str


class ErrorResponseModel(pydantic.BaseModel):
    error: str


def test__deserialize_request_model__success():
    request_kwargs = dict()
    _deserialize_request_model(
        DummyResponseModel(foo=1, bar="one"),
        request_kwargs,
        "Abort message does not matter here",
        "Whatever Subject",
    )
    assert json.loads(request_kwargs["content"]) == dict(foo=1, bar="one")
    assert request_kwargs["headers"] == {"Content-Type": "application/json"}


def test__deserialize_request_model__raises_Abort_if_request_kwargs_already_has_other_body_parts():
    with pytest.raises(Abort, match="Request was incorrectly structured"):
        _deserialize_request_model(
            DummyResponseModel(foo=1, bar="one"),
            dict(data=dict(foo=11)),
            "Abort message does not matter here",
            "Whatever Subject",
        )

    with pytest.raises(Abort, match="Request was incorrectly structured"):
        _deserialize_request_model(
            DummyResponseModel(foo=1, bar="one"),
            dict(json=dict(foo=11)),
            "Abort message does not matter here",
            "Whatever Subject",
        )

    with pytest.raises(Abort, match="Request was incorrectly structured"):
        _deserialize_request_model(
            DummyResponseModel(foo=1, bar="one"),
            dict(content=json.dumps(dict(foo=11))),
            "Abort message does not matter here",
            "Whatever Subject",
        )


def test_make_request__success(respx_mock, dummy_client):
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    respx_mock.get(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(
                foo=1,
                bar="one",
            ),
        ),
    )
    dummy_response_instance = make_request(
        client,
        req_path,
        "GET",
        expected_status=200,
        response_model_cls=DummyResponseModel,
    )
    assert isinstance(dummy_response_instance, DummyResponseModel)
    assert dummy_response_instance.foo == 1
    assert dummy_response_instance.bar == "one"


def test_make_request__raises_Abort_if_client_request_raises_exception(
    respx_mock,
    dummy_client,
):
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"
    original_error = httpx.RequestError("BOOM!")

    respx_mock.get(f"{DEFAULT_DOMAIN}{req_path}").mock(side_effect=original_error)

    with pytest.raises(
        Abort, match="There was a big problem: Communication with the API failed"
    ) as err_info:
        make_request(
            client,
            req_path,
            "GET",
            abort_message="There was a big problem",
            abort_subject="BIG PROBLEM",
        )
    assert err_info.value.subject == "BIG PROBLEM"
    assert err_info.value.log_message == ("There was an error making the request to the API")
    assert err_info.value.original_error == original_error


def test_make_request__raises_Abort_with_ownership_message_for_403_for_non_owners(
    respx_mock,
    dummy_client,
):
    client = dummy_client()
    req_path = "/fake-path"

    respx_mock.delete(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(
            httpx.codes.FORBIDDEN,
            json=dict(detail="This jabroni does not own this whingding"),
        ),
    )

    with pytest.raises(Abort, match="You do not own this resource") as err_info:
        make_request(
            client,
            req_path,
            "DELETE",
            expected_status=204,
            abort_message="There was a big problem",
            abort_subject="BIG PROBLEM",
        )
    assert err_info.value.subject == "BIG PROBLEM"
    assert "Resource could not be modified by non-owner" in err_info.value.log_message
    assert "This jabroni does not own this whingding" in err_info.value.log_message
    assert err_info.value.original_error is None


def test_make_request__raises_Abort_when_expected_status_is_not_None_and_response_does_not_match(
    respx_mock, dummy_client
):
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    respx_mock.get(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(
            httpx.codes.BAD_REQUEST,
            text="It blowed up",
        ),
    )

    with pytest.raises(
        Abort,
        match="There was a big problem: Received an error response",
    ) as err_info:
        make_request(
            client,
            req_path,
            "GET",
            expected_status=200,
            abort_message="There was a big problem",
            abort_subject="BIG PROBLEM",
        )
    assert err_info.value.subject == "BIG PROBLEM"
    assert err_info.value.log_message == ("Got an error code for request: 400: It blowed up")
    assert err_info.value.original_error is None


def test_make_request__does_not_raise_Abort_when_expected_status_is_None_and_response_has_fail_code(
    respx_mock, dummy_client
):
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    respx_mock.get(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(
            httpx.codes.BAD_REQUEST,
            json=dict(error="It blowed up"),
        ),
    )

    err = make_request(
        client,
        req_path,
        "GET",
        response_model_cls=ErrorResponseModel,
    )
    assert err.error == "It blowed up"


def test_make_request__returns_the_response_status_code_if_the_method_is_DELETE(
    respx_mock,
    dummy_client,
):
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    respx_mock.delete(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(error="It blowed up"),
        ),
    )

    assert make_request(client, req_path, "DELETE") == httpx.codes.OK


def test_make_request__returns_the_response_status_code_if_expect_response_is_False(
    respx_mock,
    dummy_client,
):
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    respx_mock.post(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(httpx.codes.BAD_REQUEST),
    )

    assert (
        make_request(
            client,
            req_path,
            "POST",
            expect_response=False,
        )
        == httpx.codes.BAD_REQUEST
    )


def test_make_request__raises_an_Abort_if_the_response_cannot_be_deserialized_with_JSON(
    respx_mock,
    dummy_client,
):
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    respx_mock.get(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            text="Not JSON, my dude",
        ),
    )

    with pytest.raises(
        Abort,
        match="There was a big problem: Response carried no data",
    ) as err_info:
        make_request(
            client,
            req_path,
            "GET",
            abort_message="There was a big problem",
            abort_subject="BIG PROBLEM",
        )
    assert err_info.value.subject == "BIG PROBLEM"
    assert err_info.value.log_message == "Failed unpacking json: Not JSON, my dude"
    assert isinstance(err_info.value.original_error, json.decoder.JSONDecodeError)


def test_make_request__returns_a_plain_dict_if_response_model_cls_is_None(
    respx_mock,
    dummy_client,
):
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    respx_mock.get(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(a=1, b=2, c=3),
        ),
    )

    assert make_request(client, req_path, "GET") == dict(a=1, b=2, c=3)


def test_make_request__raises_an_Abort_if_the_response_data_cannot_be_serialized(
    respx_mock, dummy_client
):
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    respx_mock.get(f"{DEFAULT_DOMAIN}{req_path}").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(a=1, b=2, c=3),
        ),
    )

    with pytest.raises(
        Abort,
        match="There was a big problem: Unexpected data in response",
    ) as err_info:
        make_request(
            client,
            req_path,
            "GET",
            abort_message="There was a big problem",
            abort_subject="BIG PROBLEM",
            response_model_cls=DummyResponseModel,
        )
    assert err_info.value.subject == "BIG PROBLEM"
    assert err_info.value.log_message == (
        f"Unexpected format in response data: {dict(a=1, b=2, c=3)}"
    )
    assert isinstance(err_info.value.original_error, pydantic.ValidationError)


def test_make_request__uses_request_model_instance_for_request_body_if_passed(
    respx_mock,
    dummy_client,
):
    client = dummy_client(headers={"content-type": "garbage"})
    req_path = "/fake-path"

    dummy_route = respx_mock.post(f"{DEFAULT_DOMAIN}{req_path}")
    dummy_route.mock(
        return_value=httpx.Response(
            httpx.codes.CREATED,
            json=dict(
                foo=1,
                bar="one",
            ),
        ),
    )
    dummy_response_instance = make_request(
        client,
        req_path,
        "POST",
        expected_status=201,
        response_model_cls=DummyResponseModel,
        request_model=DummyResponseModel(foo=1, bar="one"),
    )
    assert isinstance(dummy_response_instance, DummyResponseModel)
    assert dummy_response_instance.foo == 1
    assert dummy_response_instance.bar == "one"

    assert dummy_route.calls.last.request.content == json.dumps(
        dict(foo=1, bar="one"),
        separators=(",", ":"),
    ).encode("utf-8")
    assert dummy_route.calls.last.request.headers["Content-Type"] == "application/json"



import pytest

from armasec_cli.cache import (
    _get_token_paths,
    load_tokens_from_cache,
    save_tokens_to_cache,
    clear_token_cache,
    init_cache,
)
from armasec_cli.exceptions import Abort
from armasec_cli.schemas import TokenSet


@pytest.mark.usefixtures("override_cache_dir")
def test_load_tokens_from_cache__success(make_token):
    access_token = make_token(
        azp="dummy-client",
        email="good@email.com",
        expires="2022-02-16 22:30:00",
    )
    refresh_token = "dummy-refresh-token"
    (access_token_path, refresh_token_path) = _get_token_paths()
    access_token_path.write_text(access_token)
    refresh_token_path.write_text(refresh_token)

    token_set = load_tokens_from_cache()

    assert token_set.access_token == access_token
    assert token_set.refresh_token == refresh_token


@pytest.mark.usefixtures("override_cache_dir")
def test_load_tokens_from_cache__omits_refresh_token_if_it_is_not_found(make_token):
    access_token = make_token(
        azp="dummy-client",
        email="good@email.com",
        expires="2022-02-16 22:30:00",
    )
    (access_token_path, _) = _get_token_paths()
    access_token_path.write_text(access_token)

    token_set = load_tokens_from_cache()

    assert token_set.access_token == access_token
    assert token_set.refresh_token is None


@pytest.mark.usefixtures("override_cache_dir")
def test_save_tokens_to_cache__success(make_token):
    access_token = make_token(
        azp="dummy-client",
        email="good@email.com",
        expires="2022-02-16 22:30:00",
    )
    refresh_token = "dummy-refresh-token"
    (access_token_path, refresh_token_path) = _get_token_paths()
    token_set = TokenSet(
        access_token=access_token,
        refresh_token=refresh_token,
    )

    save_tokens_to_cache(token_set)

    assert access_token_path.exists()
    assert access_token_path.read_text() == access_token
    assert access_token_path.stat().st_mode & 0o777 == 0o600

    assert refresh_token_path.exists()
    assert refresh_token_path.read_text() == refresh_token
    assert access_token_path.stat().st_mode & 0o777 == 0o600


@pytest.mark.usefixtures("override_cache_dir")
def test_save_tokens_to_cache__only_saves_access_token_if_refresh_token_is_not_defined(
    make_token,
):
    access_token = make_token(
        azp="dummy-client",
        email="good@email.com",
        expires="2022-02-16 22:30:00",
    )
    (access_token_path, refresh_token_path) = _get_token_paths()
    token_set = TokenSet(
        access_token=access_token,
    )

    save_tokens_to_cache(token_set)

    assert access_token_path.exists()
    assert access_token_path.read_text() == access_token

    assert not refresh_token_path.exists()


@pytest.mark.usefixtures("override_cache_dir")
def test_clear_token_cache__success(make_token):
    access_token = make_token(
        azp="dummy-client",
        email="good@email.com",
        expires="2022-02-16 22:30:00",
    )
    refresh_token = "dummy-refresh-token"
    (access_token_path, refresh_token_path) = _get_token_paths()
    access_token_path.write_text(access_token)
    refresh_token_path.write_text(refresh_token)

    assert access_token_path.exists()
    assert refresh_token_path.exists()

    clear_token_cache()

    assert not access_token_path.exists()


@pytest.mark.usefixtures("override_cache_dir")
def test_clear_token_cache__does_not_fail_if_no_tokens_are_in_cache():
    (access_token_path, refresh_token_path) = _get_token_paths()

    assert not access_token_path.exists()
    assert not refresh_token_path.exists()

    clear_token_cache()


@pytest.mark.usefixtures("override_cache_dir")
def test_init_cache__success(tmp_path, mocker):
    new_cache_dir = tmp_path / "cache"
    with mocker.patch("armasec_cli.cache.cache_dir", new=new_cache_dir):

        @init_cache
        def _helper():
            assert new_cache_dir.exists()
            token_path = new_cache_dir / "token"
            assert token_path.exists()
            info_file_path = new_cache_dir / "info.txt"
            assert info_file_path.exists()
            assert "cache" in info_file_path.read_text()

        _helper()


@pytest.mark.usefixtures("override_cache_dir")
def test_init_cache__raises_Abort_if_cache_dir_is_not_usable(tmp_path):
    info_file_path = tmp_path / "info.txt"
    info_file_path.write_text("blah")
    info_file_path.chmod(0o000)

    @init_cache
    def _helper():
        pass

    with pytest.raises(Abort, match="check your home directory permissions"):
        _helper()



"""
These tests verify the functionality of the TokenDecoder.
"""

from uuid import uuid4
from unittest import mock

import pytest

from armasec.exceptions import AuthenticationError, PayloadMappingError
from armasec.schemas.jwks import JWK, JWKs
from armasec.token_decoder import TokenDecoder, extract_keycloak_permissions


def test_get_decode_key(rs256_jwk, build_rs256_token, rs256_kid):
    """
    Verify that the correct JWK is returned and matches on the 'kid' header.
    """
    jwks = JWKs(
        keys=[
            rs256_jwk,
            JWK(
                alg="RS256",
                kty="RSA",
                kid="unmatchable",
                n="threeve",
                e="AQAB",
            ),
        ],
    )

    decoder = TokenDecoder(jwks)
    token = build_rs256_token()
    key = decoder.get_decode_key(token)
    assert key["kid"] == rs256_jwk.kid


def test_get_decode_key__fail_if_kid_does_not_match(build_rs256_token):
    """
    Verify that an AuthenticationError is raised if no matching JWK is found.
    """
    jwks = JWKs(
        keys=[
            JWK(
                alg="RS256",
                kty="RSA",
                kid="unmatchable",
                n="threeve",
                e="AQAB",
            ),
        ],
    )
    decoder = TokenDecoder(jwks)
    token = build_rs256_token()
    with pytest.raises(AuthenticationError, match="Could not find a matching jwk"):
        decoder.get_decode_key(token)


def test_decode__success(rs256_jwk, build_rs256_token):
    """
    Verify that an RS256Decoder can successfully decode a valid jwt.
    """
    decoder = TokenDecoder(JWKs(keys=[rs256_jwk]))
    token = build_rs256_token(
        claim_overrides=dict(
            sub="test_decode-test-sub",
            azp="some-fake-id",
        ),
    )
    token_payload = decoder.decode(token)
    assert token_payload.sub == "test_decode-test-sub"
    assert token_payload.client_id == "some-fake-id"
    assert token_payload.original_token == token


def test_decode__with_token_built_with_format_keycloak(rs256_jwk, build_rs256_token):
    """
    Verify that an RS256Decoder can successfully decode a valid jwt that was encoded with the
    format_keycloak flag.
    """
    decoder = TokenDecoder(
        JWKs(keys=[rs256_jwk]),
        permission_extractor=extract_keycloak_permissions,
    )
    token = build_rs256_token(
        claim_overrides=dict(
            permissions=["read:stuff", "write:stuff"],
        ),
        format_keycloak=True,
    )
    token_payload = decoder.decode(token)
    assert token_payload.permissions == ["read:stuff", "write:stuff"]
    assert token_payload.client_id is not None
    assert token_payload.client_id.startswith("test-client")

    token = build_rs256_token(
        claim_overrides=dict(
            azp="my-client-id",
            permissions=["read:stuff", "write:stuff"],
        ),
        format_keycloak=True,
    )
    token_payload = decoder.decode(token)
    assert token_payload.permissions == ["read:stuff", "write:stuff"]
    assert token_payload.client_id == "my-client-id"


def test_decode__fails_when_jwt_decode_throws_an_error(rs256_jwk):
    """
    This test verifies that the ``decode()`` raises an exception with a helpful message when it
    fails to decode a token.
    """
    decoder = TokenDecoder(JWKs(keys=[rs256_jwk]))
    with mock.patch("jose.jwt.decode", side_effect=Exception("BOOM!")):
        with pytest.raises(AuthenticationError, match="Failed to decode token string"):
            decoder.decode("doesn't matter what token we pass here")


def test_decode__with_permission_extractor(rs256_jwk, build_rs256_token):
    """
    Verify that an RS256Decoder can extract permissions from a valid jwt.
    """
    token = build_rs256_token(
        claim_overrides=dict(
            sub="test_decode-test-sub",
            azp="some-fake-id",
            permissions=[],
            resource_access=dict(default=dict(roles=["read:stuff", "write:stuff"])),
        ),
    )

    def extractor(token_dict):
        return token_dict["resource_access"]["default"]["roles"]

    decoder = TokenDecoder(
        JWKs(keys=[rs256_jwk]),
        permission_extractor=extractor,
    )
    token_payload = decoder.decode(token)
    assert token_payload.sub == "test_decode-test-sub"
    assert token_payload.client_id == "some-fake-id"
    assert token_payload.permissions == ["read:stuff", "write:stuff"]


def test_decode__permission_extractor_raises_error(rs256_jwk, build_rs256_token):
    """
    Verify that an RS256Decoder handles a failure in the permission extractor.

    If an exception is raised by the permission extractor, it should be handled
    by the decoder and PayloadMappingError should be raised instead.
    """
    token = build_rs256_token(
        claim_overrides=dict(
            sub="test_decode-test-sub",
            azp="some-fake-id",
            permissions=[],
            resource_access=dict(default=dict(roles=["read:stuff", "write:stuff"])),
        ),
    )

    def extractor(_):
        raise RuntimeError("Boom!")

    decoder = TokenDecoder(
        JWKs(keys=[rs256_jwk]),
        permission_extractor=extractor,
    )
    with pytest.raises(
        PayloadMappingError,
        match="Failed to map decoded token.*Boom!",
    ):
        decoder.decode(token)


def test_extract_keycloak_permissions():
    """
    Verify the `extract_keycloak_permissions()` works as intended.

    It should correctly extract's the client's role as the permissions to be used in the
    TokenPayload.
    """
    client_id = uuid4()
    decoded_token = {
        "exp": 1728627701,
        "iat": 1728626801,
        "jti": "24fdb7ef-d773-4e6b-982a-b8126dd58af7",
        "sub": "dfa64115-40b5-46ab-924c-c376e73f631d",
        "azp": client_id,
        "resource_access": {
            client_id: {"roles": ["read:stuff"]},
        },
    }

    assert extract_keycloak_permissions(decoded_token) == ["read:stuff"]



import pytest

from armasec_cli.exceptions import Abort


def test_Abort_handle_errors():
    original_error = RuntimeError("Boom!")
    with pytest.raises(Abort) as err_info:
        with Abort.handle_errors("Something went wrong"):
            raise original_error

    assert err_info.value.original_error is original_error
    assert "Boom!" in str(err_info.value)
    assert "Something went wrong" in str(err_info.value)



import pendulum
import plummet
import pytest
from jose import jwt
from typer import Context

from armasec_cli.schemas import CliContext


@pytest.fixture
def make_token():
    """
    Provide a fixture that returns a helper function for creating an access_token for testing.
    """

    def _helper(azp: str = None, email: str = None, expires=plummet.AGGREGATE_TYPE) -> str:
        """
        Create an access_token with a given user email, org name, and expiration moment.
        """
        expires_moment: pendulum.DateTime = plummet.momentize(expires)

        extra_claims = dict()
        if azp is not None:
            extra_claims["azp"] = azp
        if email is not None:
            extra_claims["email"] = email

        return jwt.encode(
            {
                "exp": expires_moment.int_timestamp,
                **extra_claims,
            },
            "fake-secret",
            algorithm="HS256",
        )

    return _helper


@pytest.fixture
def override_cache_dir(tmp_path, mocker):
    with mocker.patch("armasec_cli.cache.cache_dir", new=tmp_path):
        token_path = tmp_path / "token"
        token_path.mkdir()
        yield


@pytest.fixture
def mock_context(mocker):
    typer_context = Context(mocker.MagicMock())
    typer_context.obj = CliContext()
    return typer_context



import json
from pathlib import Path

import pytest
from typer import Context

from armasec_cli.exceptions import Abort
from armasec_cli.schemas import CliContext
from armasec_cli.config import init_settings, attach_settings, dump_settings


def test_init_settings__success():
    settings = init_settings(
        oidc_domain="test.domain",
        oidc_audience="https://test.domain/api",
        oidc_client_id="test-client-id",
    )
    assert settings.oidc_domain == "test.domain"
    assert settings.oidc_audience == "https://test.domain/api"
    assert settings.oidc_client_id == "test-client-id"


def test_init_settings__raises_Abort_on_invalid_config():
    with pytest.raises(Abort):
        init_settings()


def test_dump_settings__success(tmp_path, mocker):
    dummy_settings_path = tmp_path / "settings.json"
    with mocker.patch("armasec_cli.config.settings_path", new=dummy_settings_path):
        dump_settings(
            init_settings(
                oidc_domain="test.domain",
                oidc_audience="https://test.domain/api",
                oidc_client_id="test-client-id",
            )
        )
        settings_dict = json.loads(dummy_settings_path.read_text())

    assert settings_dict["oidc_domain"] == "test.domain"
    assert settings_dict["oidc_audience"] == "https://test.domain/api"
    assert settings_dict["oidc_client_id"] == "test-client-id"


def test_attach_settings__success(tmp_path, mocker, mock_context):
    dummy_settings_path = tmp_path / "settings.json"
    with mocker.patch("armasec_cli.config.settings_path", new=dummy_settings_path):
        dump_settings(
            init_settings(
                oidc_domain="test.domain",
                oidc_audience="https://test.domain/api",
                oidc_client_id="test-client-id",
            )
        )

        @attach_settings
        def _helper(ctx):
            assert ctx.obj.settings.oidc_domain == "test.domain"
            assert ctx.obj.settings.oidc_audience == "https://test.domain/api"
            assert ctx.obj.settings.oidc_client_id == "test-client-id"

        _helper(mock_context)


def test_attach_settings__raises_Abort_if_settings_file_is_not_found(mocker):
    with mocker.patch("armasec_cli.config.settings_path", new=Path("fake-path")):

        @attach_settings
        def _helper(*_):
            pass

        typer_context = Context(mocker.MagicMock())
        typer_context.obj = CliContext()
        with pytest.raises(Abort, match="No settings file found"):
            _helper(typer_context)



import os

from armasec import Armasec
from fastapi import FastAPI, Depends


app = FastAPI()
armasec = Armasec(
    domain=os.environ.get("ARMASEC_DOMAIN"),
    audience=os.environ.get("ARMASEC_AUDIENCE"),
)

@app.get("/stuff", dependencies=[Depends(armasec.lockdown("read:stuff"))])
async def check_access():
    return dict(message="Successfully authenticated!")



from fastapi import FastAPI, Depends

from armasec_fastapi import Armasec


app = FastAPI()
armasec = Armasec(
    domain="auth0-armasec-tutorial.us.auth0.com",
    audience="https://auth0-api",
    debug_logger=print,
    debug_exceptions=True,
)

@app.get("/stuff", dependencies=[Depends(armasec.lockdown("read:stuff"))])
async def check_access():
    return dict(message="Successfully authenticated!")



import httpx
import pendulum
import plummet
import pytest
from jose import ExpiredSignatureError

from armasec_cli.auth import (
    fetch_auth_tokens,
    init_persona,
    refresh_access_token,
    validate_token_and_extract_identity,
)
from armasec_cli.cache import _get_token_paths
from armasec_cli.schemas import TokenSet, CliContext
from armasec_cli.exceptions import Abort
from armasec_cli.time_loop import Tick
from armasec_cli.config import init_settings
from armasec_cli.client import build_client


def test_validate_token_and_extract_identity__success(make_token):
    access_token = make_token(
        azp="dummy-client",
        email="good@email.com",
        expires="2022-02-16 22:30:00",
    )
    with plummet.frozen_time("2022-02-16 21:30:00"):
        identity_data = validate_token_and_extract_identity(TokenSet(access_token=access_token))
    assert identity_data.client_id == "dummy-client"
    assert identity_data.email == "good@email.com"


def test_validate_token_and_extract_identity__re_raises_ExpiredSignatureError(make_token):
    access_token = make_token(
        azp="dummy-client",
        email="good@email.com",
        expires="2022-02-16 20:30:00",
    )
    with plummet.frozen_time("2022-02-16 21:30:00"):
        with pytest.raises(ExpiredSignatureError):
            validate_token_and_extract_identity(TokenSet(access_token=access_token))


def test_validate_token_and_extract_identity__raises_abort_on_empty_token():
    test_token_set = TokenSet(access_token="")
    with pytest.raises(Abort, match="Access token file exists but it is empty"):
        validate_token_and_extract_identity(test_token_set)


def test_validate_token_and_extract_identity__raises_abort_on_unknown_error(mocker):
    test_token_set = TokenSet(access_token="BOGUS-TOKEN")
    mocker.patch("jose.jwt.decode", side_effect=Exception("BOOM!"))
    with pytest.raises(Abort, match="There was an unknown error while validating"):
        validate_token_and_extract_identity(test_token_set)


def test_validate_token_and_extract_identity__raises_abort_if_token_is_missing_identity_data(
    make_token,
):
    access_token = make_token(expires="2022-02-16 22:30:00")
    with plummet.frozen_time("2022-02-16 21:30:00"):
        with pytest.raises(Abort, match="error extracting the user's identity"):
            validate_token_and_extract_identity(TokenSet(access_token=access_token))


@pytest.mark.usefixtures("override_cache_dir")
def test_init_persona__success(make_token, mock_context):
    access_token = make_token(
        azp="dummy-client",
        email="good@email.com",
        expires="2022-02-16 22:30:00",
    )
    refresh_token = "dummy-refresh-token"
    (access_token_path, refresh_token_path) = _get_token_paths()
    access_token_path.write_text(access_token)
    refresh_token_path.write_text(refresh_token)

    with plummet.frozen_time("2022-02-16 21:30:00"):
        persona = init_persona(mock_context)

    assert persona.token_set.access_token == access_token
    assert persona.token_set.refresh_token == refresh_token
    assert persona.identity_data.client_id == "dummy-client"
    assert persona.identity_data.email == "good@email.com"


@pytest.mark.usefixtures("override_cache_dir")
def test_init_persona__uses_passed_token_set(make_token, mock_context):
    access_token = make_token(
        azp="dummy-client",
        email="good@email.com",
        expires="2022-02-16 22:30:00",
    )
    refresh_token = "dummy-refresh-token"
    (access_token_path, refresh_token_path) = _get_token_paths()

    token_set = TokenSet(
        access_token=access_token,
        refresh_token=refresh_token,
    )

    assert not access_token_path.exists()
    assert not refresh_token_path.exists()

    with plummet.frozen_time("2022-02-16 21:30:00"):
        persona = init_persona(mock_context, token_set)

    assert persona.token_set.access_token == access_token
    assert persona.token_set.refresh_token == refresh_token
    assert persona.identity_data.client_id == "dummy-client"
    assert persona.identity_data.email == "good@email.com"

    assert access_token_path.exists()
    assert access_token_path.read_text() == access_token
    assert refresh_token_path.exists()


@pytest.mark.usefixtures("override_cache_dir")
def test_init_persona__refreshes_access_token_if_it_is_expired(make_token, respx_mock):
    access_token = make_token(
        azp="dummy-client",
        email="good@email.com",
        expires="2022-02-16 22:30:00",
    )
    refresh_token = "dummy-refresh-token"
    (access_token_path, refresh_token_path) = _get_token_paths()
    access_token_path.write_text(access_token)
    refresh_token_path.write_text(refresh_token)

    refreshed_access_token = make_token(
        azp="dummy-client",
        email="good@email.com",
        expires="2022-02-17 22:30:00",
    )

    respx_mock.post("https://test.domain/protocol/openid-connect/token").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(access_token=refreshed_access_token),
        ),
    )

    settings = init_settings(
        oidc_domain="test.domain",
        oidc_audience="https://test.domain/api",
        oidc_client_id="test-client-id",
    )
    client = build_client(settings)
    cli_context = CliContext(
        settings=settings,
        client=client,
    )

    with plummet.frozen_time("2022-02-16 23:30:00"):
        persona = init_persona(cli_context)

    assert persona.token_set.access_token == refreshed_access_token
    assert persona.token_set.refresh_token == refresh_token
    assert persona.identity_data.client_id == "dummy-client"
    assert persona.identity_data.email == "good@email.com"

    assert access_token_path.exists()
    assert access_token_path.read_text() == refreshed_access_token
    assert refresh_token_path.exists()
    assert refresh_token_path.read_text() == refresh_token


def test_refresh_access_token__success(make_token, respx_mock, mock_context):
    access_token = "expired-access-token"
    refresh_token = "dummy-refresh-token"
    token_set = TokenSet(access_token=access_token, refresh_token=refresh_token)

    refreshed_access_token = make_token(
        azp="dummy-client",
        email="good@email.com",
        expires="2022-02-17 22:30:00",
    )

    respx_mock.post("https://test.domain/protocol/openid-connect/token").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(access_token=refreshed_access_token),
        ),
    )

    settings = init_settings(
        oidc_domain="test.domain",
        oidc_audience="https://test.domain/api",
        oidc_client_id="test-client-id",
    )
    client = build_client(settings)
    cli_context = CliContext(
        settings=settings,
        client=client,
    )

    refresh_access_token(cli_context, token_set)

    assert token_set.access_token == refreshed_access_token


def test_refresh_access_token__raises_abort_on_non_200_response(respx_mock):
    access_token = "expired-access-token"
    refresh_token = "dummy-refresh-token"
    token_set = TokenSet(access_token=access_token, refresh_token=refresh_token)

    respx_mock.post("https://test.domain/protocol/openid-connect/token").mock(
        return_value=httpx.Response(
            httpx.codes.BAD_REQUEST,
            json=dict(error_description="BOOM!"),
        ),
    )

    settings = init_settings(
        oidc_domain="test.domain",
        oidc_audience="https://test.domain/api",
        oidc_client_id="test-client-id",
    )
    client = build_client(settings)
    cli_context = CliContext(
        settings=settings,
        client=client,
    )

    with pytest.raises(Abort, match="auth token could not be refreshed"):
        refresh_access_token(cli_context, token_set)


def test_fetch_auth_tokens__success(respx_mock):
    access_token = "dummy-access-token"
    refresh_token = "dummy-refresh-token"
    respx_mock.post("https://test.domain/protocol/openid-connect/auth/device").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(
                device_code="dummy-code",
                verification_uri_complete="https://dummy-uri.com",
                interval=1,
            ),
        ),
    )
    respx_mock.post("https://test.domain/protocol/openid-connect/token").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(
                access_token=access_token,
                refresh_token=refresh_token,
            ),
        ),
    )

    settings = init_settings(
        oidc_domain="test.domain",
        oidc_audience="https://test.domain/api",
        oidc_client_id="test-client-id",
    )
    client = build_client(settings)
    cli_context = CliContext(
        settings=settings,
        client=client,
    )

    token_set = fetch_auth_tokens(cli_context)
    assert token_set.access_token == access_token
    assert token_set.refresh_token == refresh_token


def test_fetch_auth_tokens__raises_Abort_when_it_times_out_waiting_for_the_user(
    respx_mock,
    mocker,
):
    respx_mock.post("https://test.domain/protocol/openid-connect/auth/device").mock(
        return_value=httpx.Response(
            httpx.codes.OK,
            json=dict(
                device_code="dummy-code",
                verification_uri_complete="https://dummy-uri.com",
                interval=0,
            ),
        ),
    )
    respx_mock.post("https://test.domain/protocol/openid-connect/token").mock(
        return_value=httpx.Response(
            httpx.codes.BAD_REQUEST,
            json=dict(error="authorization_pending"),
        ),
    )

    settings = init_settings(
        oidc_domain="test.domain",
        oidc_audience="https://test.domain/api",
        oidc_client_id="test-client-id",
    )
    client = build_client(settings)
    cli_context = CliContext(
        settings=settings,
        client=client,
    )

    one_tick = Tick(
        counter=1,
        elapsed=pendulum.Duration(seconds=1),
        total_elapsed=pendulum.Duration(seconds=1),
    )
    mocker.patch("armasec_cli.auth.TimeLoop", return_value=[one_tick])
    with pytest.raises(Abort, match="not completed in time"):
        fetch_auth_tokens(cli_context)



import os
import sys

from armasec import Armasec, TokenPayload
from armasec.exceptions import ArmasecError
from armasec.pluggable import hookimpl, plugin_manager
from fastapi import FastAPI, Depends
from loguru import logger
from pydantic import BaseModel
from starlette import status

# Local data store for added "subscribers" for this demo
subscribers = set()

class Subscriber(BaseModel):
    email: str


# Custom plugin code
class PluginError(ArmasecError):
    status_code = status.HTTP_402_PAYMENT_REQUIRED
    detail = "User is not subscribed."


@hookimpl
def armasec_plugin_check(token_payload: TokenPayload):
    logger.debug("Applying check from example plugin")
    PluginError.require_condition(
        getattr(token_payload, "email") in subscribers,
        "User is not subscribed!",
    )

plugin_manager.register(sys.modules[__name__])


# FastAPI app definition
app = FastAPI()
armasec = Armasec(
    domain=os.environ.get("ARMASEC_DOMAIN"),
    audience=os.environ.get("ARMASEC_AUDIENCE"),
    payload_claim_mapping=dict(permissions="resource_access.armasec_tutorial.roles"),
    use_https=False,
    debug_logger=logger.debug,
)

@app.get("/stuff", dependencies=[Depends(armasec.lockdown("read:stuff"))])
async def check_access():
    return dict(message="Successfully authenticated!")


@app.post("/stuff", dependencies=[Depends(armasec.lockdown("read:stuff", skip_plugins=True))])
async def add_subscriber(subscriber: Subscriber):
    subscribers.add(subscriber.email)
    return dict(message=f"Added subscriber {subscriber.email}!")



"""
Verify that the TokenSecurity functions as expected with FastAPI's dependeny injection on endpoints
"""

from typing import List, Optional

import asgi_lifespan
import fastapi
import httpx
import pendulum
import pytest
import starlette
from plummet import frozen_time

from armasec.token_security import PermissionMode, TokenSecurity
from armasec.pluggable import plugin_manager, hookimpl
from armasec.exceptions import ArmasecError


@pytest.fixture
async def app():
    """
    Provides an instance of a FastAPI app that is to be used only for testing purposes.
    """
    return fastapi.FastAPI()


@pytest.fixture
async def build_secure_endpoint(app, rs256_domain_config, mock_openid_server, rs256_jwk):
    """
    Provides a method that dynamically builds a GET route on the app that requires a header with a
    valid token that includes the supplied scopes.
    """

    def _helper(
        path: str,
        scopes: Optional[List[str]] = None,
        permission_mode: PermissionMode = PermissionMode.ALL,
        **kwargs,
    ):
        """
        Adds a route onto the app at the provide path. If scopes are provided, they are supplied
        to the injected TokenSecurity that is added to the route. Uses supplied permission_mode if
        supplied
        """

        @app.get(
            path,
            dependencies=[
                fastapi.Depends(
                    TokenSecurity(
                        [rs256_domain_config],
                        scopes=scopes,
                        permission_mode=permission_mode,
                        **kwargs,
                    )
                ),
            ],
        )
        async def _():
            return dict(good="to go")

    return _helper


@pytest.fixture
async def client(app, build_secure_endpoint):
    """
    Provides a FastAPI client against which httpx requests can be made. Includes a "/secure"
    endpoint that requires auth via the TokenSecurity injectable.
    """
    build_secure_endpoint("/secure")
    async with asgi_lifespan.LifespanManager(app):
        async with httpx.AsyncClient(transport=httpx.ASGITransport(app=app), base_url="http://test") as client:
            yield client


@frozen_time("2021-09-16 20:56:00")
async def test_injector_allows_authorized_request(client, build_rs256_token):
    """
    This test verifies that access is granted to requests with valid auth headers on endpoints that
    are secured by armasec's injectable security instances.
    """
    exp = pendulum.parse("2021-09-17 20:56:00", tz="UTC")
    token = build_rs256_token(
        claim_overrides=dict(sub="me", permissions=["read:all"], exp=exp.timestamp()),
    )
    headers = {"Authorization": f"bearer {token}"}
    response = await client.get("/secure", headers=headers)
    assert response.status_code == starlette.status.HTTP_200_OK


async def test_injector_requires_correctly_encoded_token(client):
    """
    This test verifies that access is denied to requests when the jwt is invalid.
    """
    headers = {"Authorization": "bearer bad-token"}
    response = await client.get("/secure", headers=headers)
    assert response.status_code == starlette.status.HTTP_401_UNAUTHORIZED
    assert "Not authenticated" in response.text


@frozen_time("2021-09-16 20:56:00")
async def test_injector_requires_all_scopes(client, build_secure_endpoint, build_rs256_token):
    """
    This test verifies that access is granted to requests with valid auth headers where the token
    carries all of the scopes required by the endpoint.
    """
    exp = pendulum.parse("2021-09-17 20:56:00", tz="UTC")
    token = build_rs256_token(
        claim_overrides=dict(sub="me", permissions=["read:all", "read:other"], exp=exp.timestamp()),
    )
    headers = {"Authorization": f"bearer {token}"}

    build_secure_endpoint("/requires_read_all", scopes=["read:all"])
    response = await client.get("/requires_read_all", headers=headers)
    assert response.status_code == starlette.status.HTTP_200_OK

    build_secure_endpoint("/requires_read_more", scopes=["read:all", "read:more"])
    response = await client.get("/requires_read_more", headers=headers)
    assert response.status_code == starlette.status.HTTP_403_FORBIDDEN
    assert "Not authorized" in response.text


@frozen_time("2021-09-27 22:22:00")
async def test_injector_requires_some_scopes(client, build_secure_endpoint, build_rs256_token):
    """
    This test verifies that access is granted to requests with valid auth headers where the token
    carries some of the scopes required by the endpoint.
    """
    exp = pendulum.parse("2021-09-28 22:22:00", tz="UTC")
    token = build_rs256_token(
        claim_overrides=dict(sub="me", permissions=["read:all", "read:other"], exp=exp.timestamp()),
    )
    headers = {"Authorization": f"bearer {token}"}

    build_secure_endpoint(
        "/requires_read_all",
        scopes=["read:all"],
        permission_mode=PermissionMode.SOME,
    )
    response = await client.get("/requires_read_all", headers=headers)
    assert response.status_code == starlette.status.HTTP_200_OK

    token = build_rs256_token(
        claim_overrides=dict(sub="me", permissions=["read:something-else"], exp=exp.timestamp()),
    )
    headers = {"Authorization": f"bearer {token}"}
    build_secure_endpoint(
        "/requires_read_more",
        scopes=["read:all", "read:more"],
        permission_mode=PermissionMode.SOME,
    )
    response = await client.get("/requires_read_more", headers=headers)
    assert response.status_code == starlette.status.HTTP_403_FORBIDDEN
    assert "Not authorized" in response.text


@frozen_time("2021-09-27 22:22:00")
async def test_injector_raises_error_on_unknown_permission_mode(
    client, build_secure_endpoint, build_rs256_token
):
    """
    This test verifies that a request is unauthorized if an unkown permission mode is set.
    """
    exp = pendulum.parse("2021-09-28 22:22:00", tz="UTC")
    token = build_rs256_token(
        claim_overrides=dict(sub="me", permissions=["read:all"], exp=exp.timestamp()),
    )
    headers = {"Authorization": f"bearer {token}"}

    build_secure_endpoint(
        "/requires_read_all",
        scopes=["read:all"],
        permission_mode="Not a real permission",  # type: ignore
    )
    response = await client.get("/requires_read_all", headers=headers)
    assert response.status_code == starlette.status.HTTP_403_FORBIDDEN


@frozen_time("2021-09-16 20:56:00")
async def test_injector_validates_with_plugins(client, build_rs256_token):
    """
    This test verifies that a request is also validated against plugin implementations
    that have been registered.
    """

    class DummyError(ArmasecError):
        status_code = starlette.status.HTTP_418_IM_A_TEAPOT
        detail = "Ka-Boom!"

    class DummyImplementation:
        @staticmethod
        @hookimpl
        def armasec_plugin_check():
            DummyError.require_condition(allow, "Request is not allowed by dummy plugin.")

    plugin_manager.register(DummyImplementation)
    try:
        exp = pendulum.parse("2021-09-17 20:56:00", tz="UTC")
        token = build_rs256_token(
            claim_overrides=dict(sub="me", permissions=["read:all"], exp=exp.timestamp()),
        )
        headers = {"Authorization": f"bearer {token}"}

        allow = True
        response = await client.get("/secure", headers=headers)
        assert response.status_code == starlette.status.HTTP_200_OK

        allow = False
        response = await client.get("/secure", headers=headers)
        assert response.status_code == starlette.status.HTTP_418_IM_A_TEAPOT
        assert response.json()["detail"] == "Ka-Boom!"
    finally:
        plugin_manager.unregister(DummyImplementation)


@frozen_time("2021-09-16 20:56:00")
async def test_injector_skip_plugins_flag(client, build_rs256_token, build_secure_endpoint):
    """
    This test verifies that a request is not validated against plugin implementations
    that have been registered if the `skip_plugins` flag is supplied.
    """

    class DummyError(ArmasecError):
        status_code = starlette.status.HTTP_418_IM_A_TEAPOT
        detail = "Ka-Boom!"

    class DummyImplementation:
        @staticmethod
        @hookimpl
        def armasec_plugin_check():
            DummyError.require_condition(allow, "Request is not allowed by dummy plugin.")

    build_secure_endpoint("/plugin_skipper", skip_plugins=True)

    plugin_manager.register(DummyImplementation)
    try:
        exp = pendulum.parse("2021-09-17 20:56:00", tz="UTC")
        token = build_rs256_token(
            claim_overrides=dict(sub="me", permissions=["read:all"], exp=exp.timestamp()),
        )
        headers = {"Authorization": f"bearer {token}"}

        allow = False
        response = await client.get("/plugin_skipper", headers=headers)
        assert response.status_code == starlette.status.HTTP_200_OK
    finally:
        plugin_manager.unregister(DummyImplementation)



import pendulum
from plummet import frozen_time

from armasec.token_payload import TokenPayload


@frozen_time("2021-08-12 16:38:00")
def test_to_dict():
    """
    This test verifies that the `to_dict()` method produces a dictionary representation of the
    TokenPayload instance's data.
    """
    exp = pendulum.parse("2021-08-13 16:38:00", tz="UTC")
    payload = TokenPayload(
        sub="someone",
        permissions=["a", "b", "c"],
        expire=exp,
        client_id="some-fake-id",
    )
    assert payload.to_dict() == dict(
        sub="someone",
        permissions=["a", "b", "c"],
        exp=int(exp.timestamp()),
        client_id="some-fake-id",
    )


@frozen_time("2021-08-12 16:38:00")
def test_from_dict():
    """
    This test verifies that a TokenPayload instance can be correctly initialized with the data
    embedded in a properly constructed and valid jwt.
    """
    exp = pendulum.parse("2021-08-13 16:38:00", tz="UTC")
    payload = TokenPayload(
        **dict(
            sub="someone",
            permissions=["a", "b", "c"],
            exp=int(exp.timestamp()),
            azp="some-fake-id",
        ),
    )
    assert payload.sub == "someone"
    assert payload.permissions == ["a", "b", "c"]
    assert payload.expire == exp
    assert payload.client_id == "some-fake-id"



"""
This module provides a pytest plugin for testing.
"""

from collections import namedtuple
from contextlib import _GeneratorContextManager, contextmanager
from datetime import datetime, timezone
from typing import Callable, Optional
from uuid import uuid4

import httpx
import pytest
import respx
import starlette
from jose import jwt
from snick import dedent, strip_whitespace

from armasec.openid_config_loader import OpenidConfigLoader
from armasec.schemas.armasec_config import DomainConfig
from armasec.schemas.jwks import JWK, JWKs
from armasec.schemas.openid_config import OpenidConfig


@pytest.fixture()
def rs256_domain():
    """
    Provide a fixture that returns a domain for use in other fixtures.

    The value here doesn't really have anything to do with an actual domain name.
    """
    return "armasec.dev"


@pytest.fixture()
def rs256_domain_config(rs256_domain):
    """
    Provide a fixture that returns the DomainConfig model for the default rs256 domain.

    Args:
        rs256_domain: An implicit fixture parameter.
    """
    return DomainConfig(domain=rs256_domain, audience="https://this.api")


@pytest.fixture()
def rs256_iss(rs256_domain):
    """
    Provide a fixture that returns an issuer claim for use in other fixtures.

    Args:
        rs256_domain: An implicit fixture parameter.
    """
    return f"https://{rs256_domain}"


@pytest.fixture()
def rs256_kid():
    """
    Provide a fixture that returns a KID header value for use in other fixtures.
    """
    return "SAMPLE_KID"


@pytest.fixture()
def rs256_sub():
    """
    Provide a fixture that returns a sum claim for use in other fixtures.
    """
    return "SAMPLE_SUB"


@pytest.fixture()
def rs256_private_key():
    """
    Provide a fixture that returns a pre-generated private key for RS256 hashing in other fixtures.
    """
    return dedent(
        """
        -----BEGIN RSA PRIVATE KEY-----
        MIIEpAIBAAKCAQEAw408+QDZ10idz4ytJtwFQE4YgmrjvCoEXjtTUWQ3H4nWAAYQ
        +oE9xpr/gosNiFMuyRburvXT+Rkq8ry8tWoUzN2zViaarot+Tt9I71sVlnIsbtDZ
        +XrteMvBwjARn/MEAQEwDLvVzrBnAZrOTwrIkznyJttZh7STrt6y5X91i2MMm3xu
        9QK90kpu3rymAyT5V+AEIRzZai/ZT4YfLDutXulOVlWPQ55Xww1mbheGQ99fUMo5
        LmkxM5Jsz8ulIVvq/G/8guiKwAPJN/8S34NbkgL5GoeXT8uNDkbhtkLh5+o2T4EL
        9/ODKHqx46pHgUmBiC6wNv6uJXdH7qpaqhPR3QIDAQABAoIBABeyl/788Wk7bZRn
        UdxxsVk3nZTAa1S0Ks9YlSI56MwzofFiys/wtZHJ2sjxHPS2T+cilk4xkDyRpjjA
        UoYRku+4tjDsgLZCRU49lNMc0KLotyW+vYuUMA8BcjucI6akhomwoSgJ40Em83So
        U/QUNHZTAVtgHZtqcLMyXa+eIJqBcfsMHFkCgSF8LSD/XkRBMm1SREswDw6KqQQ0
        sZ/8TVF9sJTi3/OG8m5OfI+44AYDaMH5wKoOBcR3FBln+dEutB6JuRjmpnEjQpIT
        DggULc+Dzb/c75yhT1qZSEL3Z99JQTbytPm6boNGKmzUE9HCoY84wKfnhUDocFKW
        jnHMmKkCgYEA7/gRAtLjbJW1rbxw8xN3cyOZEJsMFt4mXMmne6nDTttVKb0wUuXJ
        H8prKAXDOzadAZgPeGJXVSNgGoNeNtkmEKDtysrRiZbWiTRxYPE36MrHFGOywjTn
        tP8qMJHmHYkxS16nqrOl0znUWv6Q6/qwd59Utuu4IJF/CxqP3Z6HPssCgYEA0J2M
        1gRgGj8NnGoIKS58gc3Aa5RdqWKoeiyXeN/zRDfMCKpPsVykvZJb4cLcEdcwe9kC
        3xpgIPaTZCPwhJ1rYiZ0/Xr7oIf0E66IeEKs/bchKcT9+sSaWgc5/zQ7aQ/XpwzU
        nKCTTeMGFUyulCIkoe2tLEQ+Mw1OphIXv17fNPcCgYEAjWgVxh81ivgRjiZ8PJEd
        E4lHmmRzVEpmOslN225nO+G9ppHolwD3ardiO7xhllQRYy4S97KjmfT1ncoJy7Jc
        XvImDhlELprnIwT3RtP+STys4ZP6c7yvSZYPa32eJ4t/s9U8YjfooLb0LwbRqW0Z
        bfRC/GOdJfv27Dkjy8muEs8CgYEAo+oHDOonMLg2U54kh2cVQVCPTngnF76DLmv3
        IGym0gUddfmL4Iowjxt+wma/T+1LFSSwUuiAe6YCrX5nr2uZQmeBKOIG8F2idAyB
        Ai0xi7Dmh9FW1kDAHtjqwxEhVS2zfnhgXij1VQ96aiX0TkR9kBYWKV/9l1NvZqF0
        s1MyAoUCgYA0wfJrCTXdWitkyfxApcmoTxt0ljqUwO6F5fhojf8PU1ouglgkRXtm
        1rSDGp7YUfODhWNSsN2P/eaDybcZo+TGtLQJ5Bai3Qxqh8xPaKCsSZbcPRLRP0w5
        CbTvFEyj6EBEH+TJL/Loa4hKFuAk7ErBAtzMCw6LchTjB/OF+dUusA==
        -----END RSA PRIVATE KEY-----
        """
    ).encode("utf-8")


@pytest.fixture()
def rs256_public_key():
    """
    Provide a fixture that returns a pre-generated public key for RS256 hashing in other fixtures.
    """
    return dedent(
        """
        -----BEGIN RSA PUBLIC KEY-----
        MIIBCgKCAQEAw408+QDZ10idz4ytJtwFQE4YgmrjvCoEXjtTUWQ3H4nWAAYQ+oE9
        xpr/gosNiFMuyRburvXT+Rkq8ry8tWoUzN2zViaarot+Tt9I71sVlnIsbtDZ+Xrt
        eMvBwjARn/MEAQEwDLvVzrBnAZrOTwrIkznyJttZh7STrt6y5X91i2MMm3xu9QK9
        0kpu3rymAyT5V+AEIRzZai/ZT4YfLDutXulOVlWPQ55Xww1mbheGQ99fUMo5Lmkx
        M5Jsz8ulIVvq/G/8guiKwAPJN/8S34NbkgL5GoeXT8uNDkbhtkLh5+o2T4EL9/OD
        KHqx46pHgUmBiC6wNv6uJXdH7qpaqhPR3QIDAQAB
        -----END RSA PUBLIC KEY-----
        """
    ).encode("utf-8")


@pytest.fixture()
def rs256_jwk(rs256_kid):
    """
    Provide a fixture that returns a JWK for use in other fixtures.

    Args:
        rs256_kid: An implicit fixture parameter.
    """
    return JWK(
        alg="RS256",
        kty="RSA",
        kid=rs256_kid,
        n=strip_whitespace(
            """
                w408-QDZ10idz4ytJtwFQE4YgmrjvCoEXjtTUWQ3H4nWAAYQ-oE9xpr_gosNiFMuyRburvXT-Rkq8ry8tWoU
                zN2zViaarot-Tt9I71sVlnIsbtDZ-XrteMvBwjARn_MEAQEwDLvVzrBnAZrOTwrIkznyJttZh7STrt6y5X91
                i2MMm3xu9QK90kpu3rymAyT5V-AEIRzZai_ZT4YfLDutXulOVlWPQ55Xww1mbheGQ99fUMo5LmkxM5Jsz8ul
                IVvq_G_8guiKwAPJN_8S34NbkgL5GoeXT8uNDkbhtkLh5-o2T4EL9_ODKHqx46pHgUmBiC6wNv6uJXdH7qpa
                qhPR3Q
            """
        ),
        e="AQAB",
    )


@pytest.fixture()
def build_rs256_token(rs256_private_key, rs256_iss, rs256_sub, rs256_kid):
    """
    Provide a fixture that returns a helper method that can build a JWT.

    The JWT is signed with the private key provided by the rs256_private_key.

    Args:
        rs256_private_key: An implicit fixture parameter.
        rs256_iss:         An implicit fixture parameter.
        rs256_sub:         An implicit fixture parameter.
    """
    base_claims = dict(
        iss=rs256_iss,
        sub=rs256_sub,
    )
    base_headers = dict(kid=rs256_kid)

    def _helper(
        claim_overrides: Optional[dict] = None,
        headers_overrides: Optional[dict] = None,
        format_keycloak: bool = False,
    ):
        """
        Encode a jwt token with the default claims and headers overridden with user supplied values.

        Args:
            claim_overrides:  A dictionary of claims to add to the token.
                              Will override any existing values upon collision
            header_overrides: A dictionary of headers to add to the token.
                              Will override any existing values upon collision
            format_keycloak:  If set, will remap "permissions" provided as a part of the
                              claim_overrides to the expected position for keycloak. If an "azp"
                              claim is not provided in the claim_overrides, it will generate
                              a random test client_id in the "azp" claim that matches the keycloak
                              structure. Example keycloak structure:

                              ```
                              {
                                "exp": 1728627701,
                                "iat": 1728626801,
                                "jti": "24fdb7ef-d773-4e6b-982a-b8126dd58af7",
                                "sub": "dfa64115-40b5-46ab-924c-c376e73f631d",
                                "azp": "my-client",
                                "resource_access": {
                                  "my-client": {
                                    "roles": [
                                      "read:stuff"
                                    ]
                                  },
                                },
                              }
                              ```

        """
        if claim_overrides is None:
            claim_overrides = dict()

        if headers_overrides is None:
            headers_overrides = dict()

        now = int(datetime.now(timezone.utc).timestamp())

        if format_keycloak and "permissions" in claim_overrides:
            test_client = claim_overrides.get("azp", f"test-client-{uuid4()}")
            claim_overrides["azp"] = test_client
            claim_overrides["resource_access"] = {
                test_client: {
                    "roles": claim_overrides.pop("permissions"),
                }
            }

        return jwt.encode(
            {
                "iat": now,
                "exp": now + 60 * 60,  # expires in an hour from now
                **base_claims,
                **claim_overrides,
            },
            rs256_private_key,
            algorithm="RS256",
            headers={
                **base_headers,
                **headers_overrides,
            },
        )

    return _helper


@pytest.fixture
def rs256_jwks_uri(rs256_domain):
    """
    Provide a fixture that returns a jwks uri for use in other fixtures.

    Args:
        rs256_jwks_uri: An implicit fixture parameter.
    """
    return f"https://{rs256_domain}/.well-known/jwks.json"


@pytest.fixture
def rs256_openid_config(rs256_iss, rs256_jwks_uri):
    """
    Provide a fixture that returns an openid configuration for use in other fixtures.

    Args:
        rs256_iss:      An implicit fixture parameter.
        rs256_jwks_uri: An implicit fixture parameter.
    """
    return OpenidConfig(
        issuer=rs256_iss,
        jwks_uri=rs256_jwks_uri,
    )


def build_mock_openid_server(
    domain, openid_config, jwk, jwks_uri
) -> Callable[[str, OpenidConfig, JWK, str], _GeneratorContextManager]:
    """
    Provide a fixture that returns a context manager that mocks opend-config routes.

    Args:
        domain:        The domain of the openid server to mock.
        openid_config: The config to return from the mocked config route.
        jwk:           The jwk to return from the mocked jwk route.
        jwks_uri:      The URL of the jwks route to mock.

    Returns:
        A context manager that, while active, mocks the openid routes needed by Armasec.
    """

    @contextmanager
    def _helper(
        domain: str = domain,
        openid_config: OpenidConfig = openid_config,
        jwk: JWK = jwk,
        jwks_uri: str = jwks_uri,
    ):
        MockOpenidRoutes = namedtuple("MockOpenidRoutes", ["openid_config_route", "jwks_route"])
        with respx.mock:
            openid_config_route = respx.get(
                OpenidConfigLoader.build_openid_config_url(domain),
            )
            openid_config_route.return_value = httpx.Response(
                starlette.status.HTTP_200_OK,
                json=openid_config.model_dump(mode="json"),
            )

            jwks = JWKs(keys=[jwk])
            jwks_route = respx.get(jwks_uri)
            jwks_route.return_value = httpx.Response(
                starlette.status.HTTP_200_OK,
                json=jwks.model_dump(mode="json"),
            )
            yield MockOpenidRoutes(openid_config_route, jwks_route)

    return _helper


@pytest.fixture
def mock_openid_server(rs256_domain, rs256_openid_config, rs256_jwk, rs256_jwks_uri):
    """
    Provide a fixture that mocks an openid server using the extension fixtures.

    Args:
        rs256_domain:        An implicit fixture parameter.
        rs256_openid_config: An implicit fixture parameter.
        rs256_jwk:           An implicit fixture parameter.
        rs256_jwks_uri:      An implicit fixture parameter.
    """
    builder = build_mock_openid_server(rs256_domain, rs256_openid_config, rs256_jwk, rs256_jwks_uri)
    with builder() as constructed_builder:
        yield constructed_builder
